package net.knarcraft.placeholdersigns;

import me.clip.placeholderapi.PlaceholderAPI;
import net.knarcraft.placeholdersigns.command.EditSignCommand;
import net.knarcraft.placeholdersigns.command.ViewSignCommand;
import net.knarcraft.placeholdersigns.container.PlaceholderSign;
import net.knarcraft.placeholdersigns.container.SignLineChangeRequest;
import net.knarcraft.placeholdersigns.handler.PlaceholderSignHandler;
import net.knarcraft.placeholdersigns.listener.SignBreakListener;
import net.knarcraft.placeholdersigns.listener.SignClickListener;
import net.knarcraft.placeholdersigns.listener.SignTextListener;
import net.knarcraft.placeholdersigns.util.ColorHelper;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.block.Sign;
import org.bukkit.block.sign.Side;
import org.bukkit.block.sign.SignSide;
import org.bukkit.command.PluginCommand;
import org.bukkit.entity.Player;
import org.bukkit.plugin.java.JavaPlugin;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;

/**
 * This plugin's main class
 */
public final class PlaceholderSigns extends JavaPlugin {

    private static PlaceholderSigns instance;
    private PlaceholderSignHandler signHandler;
    private Map<Player, SignLineChangeRequest> signChangeRequests;
    private Map<Player, Void> signViewRequests;

    /**
     * Gets an instance of this plugin
     *
     * @return <p>A plugin instance</p>
     */
    public static @NotNull PlaceholderSigns getInstance() {
        return instance;
    }

    /**
     * Gets this instance's placeholder sign handler
     *
     * @return <p>The sign handler</p>
     */
    public @NotNull PlaceholderSignHandler getSignHandler() {
        return this.signHandler;
    }

    /**
     * Registers a sign change request
     *
     * <p>A sign change request is basically the result of running the editSign command, which must be stored until the
     * player clicks a sign.</p>
     *
     * @param request <p>The sign change request to register</p>
     */
    public void addSignChangeRequest(@NotNull SignLineChangeRequest request) {
        signChangeRequests.put(request.player(), request);
    }

    /**
     * Gets a sign change request
     *
     * @param player <p>The player to get the request for</p>
     * @return <p>The sign change request, or null if not found</p>
     */
    public @Nullable SignLineChangeRequest getSignChangeRequest(@NotNull Player player) {
        return signChangeRequests.remove(player);
    }

    /**
     * Registers a sign view request
     *
     * @param player <p>The player requesting to view a sign</p>
     */
    public void addSignViewRequest(@NotNull Player player) {
        signViewRequests.put(player, null);
    }

    /**
     * Checks whether the given player has a sign view request
     *
     * @param player <p>The player to check</p>
     * @return <p>True if the player has requested to view a sign</p>
     */
    public boolean hasSignViewRequest(@NotNull Player player) {
        if (signViewRequests.containsKey(player)) {
            signViewRequests.remove(player);
            return true;
        } else {
            return false;
        }
    }

    @Override
    public void onEnable() {
        instance = this;
        signHandler = new PlaceholderSignHandler();
        signChangeRequests = new HashMap<>();
        signViewRequests = new HashMap<>();
        signHandler.load();

        if (Bukkit.getPluginManager().getPlugin("PlaceholderAPI") == null) {
            getLogger().log(Level.WARNING, "Could not find PlaceholderAPI! This plugin is required.");
            Bukkit.getPluginManager().disablePlugin(this);
            return;
        }

        // Update signs' placeholders every second
        Bukkit.getScheduler().runTaskTimer(this, this::updateSigns, 20 * 10, 20 * 5);

        Bukkit.getPluginManager().registerEvents(new SignBreakListener(), this);
        Bukkit.getPluginManager().registerEvents(new SignTextListener(), this);
        Bukkit.getPluginManager().registerEvents(new SignClickListener(), this);

        PluginCommand editCommand = Bukkit.getPluginCommand("editSign");
        if (editCommand != null) {
            editCommand.setExecutor(new EditSignCommand());
        }

        PluginCommand viewCommand = Bukkit.getPluginCommand("viewSign");
        if (viewCommand != null) {
            viewCommand.setExecutor(new ViewSignCommand());
        }
    }

    @Override
    public void onDisable() {
        // Plugin shutdown logic
    }

    /**
     * Updates all loaded and registered placeholder signs
     */
    private void updateSigns() {
        for (PlaceholderSign placeholderSign : signHandler.getSigns()) {
            // Ignore signs away from players
            Location location = placeholderSign.location();
            if (!location.getChunk().isLoaded()) {
                continue;
            }

            // If no longer a sign, remove
            if (!(location.getBlock().getState() instanceof Sign sign)) {
                signHandler.unregisterSign(placeholderSign);
                continue;
            }

            // Update placeholders
            SignSide front = sign.getSide(Side.FRONT);
            SignSide back = sign.getSide(Side.BACK);
            Map<Side, Map<Integer, String>> placeholders = placeholderSign.placeholders();
            String[] frontLines = front.getLines();
            String[] backLines = back.getLines();

            // Only update the sign if the text has changed
            boolean updateNecessary = false;
            if (placeholders.get(Side.FRONT) != null) {
                updateNecessary |= updatePlaceholders(frontLines, placeholders.get(Side.FRONT), front);
            }
            if (placeholders.get(Side.BACK) != null) {
                updateNecessary |= updatePlaceholders(backLines, placeholders.get(Side.BACK), back);
            }
            if (updateNecessary) {
                sign.update();
            }
        }
    }

    /**
     * Updates the values of placeholders on a sign
     *
     * @param lines        <p>The sign's current lines</p>
     * @param placeholders <p>The sign's original placeholder lines</p>
     * @param signSide     <p>The side of the sign to update placeholders for</p>
     * @return <p>True if text has been changed, and the sign needs to be updated</p>
     */
    private boolean updatePlaceholders(@NotNull String[] lines, @NotNull Map<Integer, String> placeholders,
                                       @NotNull SignSide signSide) {
        boolean changed = false;
        for (int i = 0; i < lines.length; i++) {
            String oldText = signSide.getLine(i);

            // The new text of the sign is either the same, or the original placeholder
            String newText;
            if (!placeholders.containsKey(i) || placeholders.get(i) == null) {
                newText = oldText;
            } else {
                newText = PlaceholderAPI.setPlaceholders(null, placeholders.get(i));
            }

            // Convert color codes
            newText = ColorHelper.translateAllColorCodes(newText);

            // Only change the line if the text has changed
            if (!newText.equals(oldText)) {
                signSide.setLine(i, newText);
                changed = true;
            }
        }

        return changed;
    }


}
