package net.knarcraft.placeholdersigns.listener;

import net.knarcraft.placeholdersigns.PlaceholderSigns;
import net.knarcraft.placeholdersigns.container.SignLineChangeRequest;
import net.knarcraft.placeholdersigns.util.ColorHelper;
import net.md_5.bungee.api.ChatColor;
import org.bukkit.Bukkit;
import org.bukkit.block.Sign;
import org.bukkit.block.sign.Side;
import org.bukkit.block.sign.SignSide;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.SignChangeEvent;
import org.bukkit.event.player.PlayerInteractEvent;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;

/**
 * A listener for placeholder signs being clicked
 */
public class SignClickListener implements Listener {

    @EventHandler(priority = EventPriority.MONITOR)
    public void onSignClick(@NotNull PlayerInteractEvent event) {
        // Ignore if not a clicked sign
        if (!event.hasBlock() || event.getClickedBlock() == null ||
                !(event.getClickedBlock().getState() instanceof Sign sign)) {
            return;
        }

        Player player = event.getPlayer();
        boolean hasSignViewRequest = PlaceholderSigns.getInstance().hasSignViewRequest(player);
        if (hasSignViewRequest) {
            printSign(sign, player);
        }

        if (sign.isWaxed() && !player.hasPermission("placeholdersigns.edit.bypass-waxed")) {
            player.sendMessage(ChatColor.RED + "You do not have the necessary permissions to edit a waxed sign.");
            return;
        }

        // Check if the player has run the /editSign command
        SignLineChangeRequest request = PlaceholderSigns.getInstance().getSignChangeRequest(player);
        if (request != null) {
            SignSide standingOn = sign.getTargetSide(player);
            Side side = sign.getSide(Side.FRONT).equals(standingOn) ? Side.FRONT : Side.BACK;
            doSignChange(sign, request, event, player, side);
        }
    }

    /**
     * Prints the current contents of a sign to a player
     *
     * @param sign   <p>The sign to print</p>
     * @param player <p>The player to display the contents to</p>
     */
    private void printSign(@NotNull Sign sign, @NotNull Player player) {
        StringBuilder builder = new StringBuilder();
        getSignText(sign.getSide(Side.FRONT), "Front:", builder);
        getSignText(sign.getSide(Side.BACK), "Back:", builder);
        player.sendMessage(builder.toString());
    }

    /**
     * Gets text from a sign side, and appends it to the given string builder
     *
     * @param signSide <p>The sign side to get text from</p>
     * @param header   <p>The header to display before printing the sign test</p>
     * @param builder  <p>The string builder to append the text to</p>
     */
    private void getSignText(@NotNull SignSide signSide, @NotNull String header, @NotNull StringBuilder builder) {
        builder.append(header).append("\n");
        for (int i = 0; i < 4; i++) {
            builder.append(signSide.getLine(i)).append("\n");
        }
    }

    /**
     * Perform the changing of a sign according to the given sign change request
     *
     * @param sign    <p>The sign to be changed</p>
     * @param request <p>The sign line change request to perform</p>
     * @param event   <p>The interaction event that triggered this</p>
     * @param player  <p>The player triggering the sign change</p>
     */
    private void doSignChange(@NotNull Sign sign, @NotNull SignLineChangeRequest request,
                              @NotNull PlayerInteractEvent event, @NotNull Player player, @NotNull Side side) {
        // Cancel the event to prevent vanilla behavior
        event.setCancelled(true);

        SignSide signSide = sign.getSide(side);
        String[] lines = signSide.getLines();

        lines[request.line()] = request.text();

        // Run the sign change event to allow protection plugins to cancel, and allow the sign text listener to trigger
        SignChangeEvent changeEvent = new SignChangeEvent(Objects.requireNonNull(event.getClickedBlock()),
                player, lines, side);
        Bukkit.getPluginManager().callEvent(changeEvent);
        if (changeEvent.isCancelled()) {
            return;
        }

        // Update the sign with the new text
        String[] finalLines = changeEvent.getLines();
        for (int i = 0; i < finalLines.length; i++) {
            signSide.setLine(i, ColorHelper.translateAllColorCodes(finalLines[i]));
        }
        sign.update();

        player.sendMessage("The sign line was successfully changed.");
    }

}
